 /*****************************************************************************
 *
 * Copyright 1999 Digi International (www.digi.com)
 *    Zhong Deng <Zhong_Deng@digi.com>
 *    Jeff Randall <Jeff_Randall@digi.com>
 *    Bob Rubendunst <Robert_Rubendunst@digi.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **************************************************************************** */

/* %W% %S% cdetherd.c Copyright (C) 1997-99 Digi International */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <sys/termios.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/syslog.h>
#include <linux/serial.h>
#include <netinet/in.h>
#include "ctio.h"
#include "fas.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define ELS_TCP_PORT    10001
/* MAXSLEEPTIME is the maximum sleep time in seconds to wait between
   ELS connection attempts. */
#define MAXSLEEPTIME 120
#define MAXDEVNAMELEN 7

/* "where" for log messages.
SYSCON - send to system console in addition to log file. */
#define SYSCON  0x2
#define NOSTAMP 0x4		/* don't print time & datestamp */

/* DEFINE max path length.
the rest of the path is /tmp/ + .older + null byte */
#define RESTOFPATH 12

#undef MAX
#define MAX(a, b) (((a) > (b)) ? (a) : (b))

void els_terminate();
void bitmash();
void err_exit();
int make_socket_connection();

extern  char    *optarg;        /* for getopt function */

/* Globals */
char Logfilename[MAXDEVNAMELEN+RESTOFPATH]; 
int log_fd = 0;
int sock_fd;
int els_fd;
long curr_time, prev_time;
/* struct ct_unit unit; */
struct sockaddr_in  cli_addr;
struct timeval sock_time;
char data_buf[4096];
char log_file[40];
els_config els_cfg;

void els_closelog();
void els_log_err(int where, const char *format, ...);
void els_log_msg(int where, char *msg);
void els_log_renamer(char *unit);
void els_terminate();
int sock_read(int sock_fd,unsigned char *buf, int wanted);

req_unit_unlock unlock_req;
rsp_unit_unlock unlock_rsp;
req_unit_inquiry inquiry_req;
rsp_unit_inquiry inquiry_rsp;
unsigned char inquiry_data[128];

/* deletes the newname file, and renames oldname to newname
Returns 0 if all went well, or errno if not.
 */
int name_shuffle(char *oldname, char *newname)
{
  struct stat buf;
  if (!stat(newname, &buf))
    unlink(newname);
  if (!stat(oldname, &buf)) {
    if (!link(oldname, newname)) {
      return unlink(oldname);
    }
  }
  return errno;
}
/* delete (unit).older, if it exits.
rename (unit).old to (unit).older
rename (unit).log to (unit).old
remove (unit).log.
*/
void els_log_renamer(char *unit)
{
  char tmpnam[] = "//tmp//";
  char name2[MAXDEVNAMELEN+RESTOFPATH];
  char name1[MAXDEVNAMELEN+RESTOFPATH];
  struct stat buf;

  strcpy(name2,tmpnam);
  strcat(name2, unit);
  strcat(name2, ".older");	/* +7 */
  strcpy(name1,tmpnam);
  strcat(name1, unit);
  strcat(name1, ".old");
  name_shuffle(name1, name2);
  strcpy(name2,tmpnam);
  strcat(name2, unit);
  strcat(name2, ".log");
  name_shuffle(name2, name1);
  strcpy(Logfilename, name2);	/*  store this Global */
  if (!stat(Logfilename, &buf))
    unlink(Logfilename);
}

/* close the log file */
void els_closelog()
{
if (log_fd)
  close(log_fd);
log_fd = 0;
}

/* common routine to print the timestamp.
Routine 0 if all went well, errno if log file could not
be opened. */

void print_timestamp()
{
  time_t cur_time;
  struct tm *tp;
  char timout[40];
  
  /*
   * Print the datestamp to the log file;
   */

  cur_time = time(0);
  tp = localtime(&cur_time);
  sprintf(timout, "%02d/%02d/%02d %02d:%02d:%02d: ",
		  tp->tm_mon+1, tp->tm_mday, tp->tm_year%100,
		  tp->tm_hour, tp->tm_min, tp->tm_sec);


 fprintf(stderr,timout);
}

void els_log_msg(int where, char *msg)
{
  if (!(where&NOSTAMP))
    print_timestamp();
  fprintf(stderr,msg);
  fflush(stderr);
  /* It may go to the system console/syslog as well. */
  if (where&SYSCON) {
    openlog("cdetherd", LOG_PID, LOG_DAEMON);
    syslog(LOG_ERR, msg);
    closelog();
  }

}

/* print standard error msg to log file.
 */
void els_log_err(int where, const char *fmt, ...)
{
  va_list vaptr;

  va_start(vaptr, fmt);
  if (!(where&NOSTAMP))
      print_timestamp();
  vfprintf(stderr,fmt,vaptr);
  va_end(vaptr);
  fflush(stderr);
  /* It may go to the system console/syslog as well. */
  if (where&SYSCON) {
    char buff[120];
    va_start(vaptr, fmt);
    vsprintf(buff,fmt,vaptr);
    va_end(vaptr);
    openlog("cdetherd", LOG_PID, LOG_DAEMON); /* syslog has datestamp */
    syslog(LOG_ERR, buff);
    closelog();
  }
}
int main(int argc, char **argv)
{
  int n, pid;
  char dev_name[20];
  char log_buf[80];
  int max_fds, nfds;
  fd_set rd_fdset;
  int first_conn = 1;
  unsigned char *ptr;
  int sleeptime = 2;		/* used to prevent syslog from filling up */
#ifdef DBG
  int i;
#endif

  els_log_renamer(argc >=2 ? argv[1] : "els");
  /* Remap stderr to log_fd. */
  if ((log_fd = open(Logfilename, O_WRONLY | O_CREAT, (mode_t)0644)) < 0) {
	perror(Logfilename);
	exit(1);
  }
  dup2(log_fd, STDERR_FILENO);

  if (argc != 3)  {
    printf("Usage: %s device ip_addr\n", argv[0]);
    els_log_err(SYSCON|NOSTAMP, "Usage: %s device ip_addr\n", argv[0]);
    els_closelog();
    exit(1);
  }

  signal (SIGHUP, SIG_IGN);
  signal (SIGINT, SIG_IGN);
  signal (SIGQUIT, SIG_IGN);
  signal (SIGTERM, els_terminate);
  
  if ((pid = fork()) < 0) {
    els_log_msg(SYSCON, "Cannot fork child process\n");
    els_closelog();
    exit(1);
  }
  else if (pid > 0)
    exit(0);

  close(0);
  close(1);
/*   close(2); */
  
  chdir("/");
  umask(0);

  els_log_msg(SYSCON, "EtherLite daemon running\n");
 again:
  sprintf(dev_name, "/dev/%s", argv[1]);
  
  if((els_fd = open(dev_name, O_RDWR)) < 0)  {
   els_log_err(SYSCON, "open ELS(%s):%s\n", dev_name,strerror(errno));
   /* if device isn't there, check less frequently for it. */
   sleep(sleeptime);
   if (sleeptime < MAXSLEEPTIME)
     sleeptime += 2;
   goto again;
 }
   sleeptime = 2;		/* reset on success */
  els_cfg.netaddr = inet_addr(argv[2]);

main_loop:
  while (make_socket_connection() < 0) {
    /* loop until connection is made */
    sleep(10);
  }
  
  if (first_conn) {
    ptr = inquiry_data + 16;		/* start of inquiry_data */
    strcpy(els_cfg.product_id, (char *) ptr + strlen((char *) ptr) + 1);
    strcpy(els_cfg.fw_ver, (char *) ptr + strlen((char *) ptr) 
	   + strlen(els_cfg.product_id) + 2);
    sprintf(log_buf, "Vendor ID: %s, Firmware Version: %s\n", 
	    ptr, els_cfg.fw_ver);
    els_log_msg(SYSCON, log_buf);
    sprintf(log_buf, "Model Name: %s, IP Address: %s\n", 
	    els_cfg.product_id, argv[2]);
    els_log_msg(SYSCON|NOSTAMP, log_buf);
    if (ioctl(els_fd, ELSADDUNIT, &els_cfg) < 0) {
      els_log_err(SYSCON, "ELSADDUNIT:%s\n",strerror(errno));
      err_exit();
    }

    first_conn = 0;
    }
  
  els_log_msg(SYSCON, "Unit Online\n");
  if (ioctl(els_fd, ELSONLINE, 0) < 0) {
    els_log_err(SYSCON, "ELSONLINE:%s\n",strerror(errno));
    err_exit();
  }

  sock_time.tv_sec = 120;
  sock_time.tv_usec = 0;
  curr_time = time(0);
  max_fds = MAX(sock_fd, els_fd) + 1;
  
  while (1)  {
    FD_ZERO(&rd_fdset);
    FD_SET(sock_fd, &rd_fdset);
    FD_SET(els_fd, &rd_fdset);
   if ((nfds = select(max_fds, &rd_fdset, (fd_set *)0, (fd_set *)0,
		       &sock_time)) < 0) {
      els_log_err(SYSCON, "select:%s\n", strerror(errno));
      break;
    }
    if (nfds == 0)  {
      els_log_msg(SYSCON, "select timed out\n");
      break;
    }
    prev_time = curr_time;
    curr_time = time(0);

    if (FD_ISSET(els_fd, &rd_fdset))  {
    /* N.B. - Linux's select already adjusts tv.sec for us, so
       we don't need to adjust it the way we do for other O/Ses -rpr */
       /*  sock_time.tv_sec -= (curr_time - prev_time); */
      if ((n = read(els_fd, data_buf, 4096)) > 0)  {
#ifdef DBG
	els_log_msg(SYSCON, "data received from els: ");
	ptr = (unsigned char *) data_buf;
	strcpy(log_buf, "    ");
	for (i = 0; i < n; i++) {
	  sprintf(log_buf + 4 + (i % 16) * 3, "%02x ", *ptr++);
	  if ((i % 16) == 15 || i == n - 1) {
	    strcat(log_buf,"\n");
	    els_log_msg(SYSCON|NOSTAMP, log_buf);
	  }
	}
#endif
	if (write(sock_fd, data_buf, n) < 0)  {
	  els_log_err(SYSCON, "write on socket:%s\n",strerror(errno));
	  break;
	}
      } else {
	els_log_err(SYSCON, "read on EtherLite:%s\n",strerror(errno));
	err_exit();
      }
    }
    if (FD_ISSET(sock_fd, &rd_fdset)) {
      sock_time.tv_sec = 120;
      if ((n = read(sock_fd, data_buf, 4096)) > 0) {
#ifdef DBG
	els_log_msg(SYSCON, "data received from socket: ");
	ptr = (unsigned char *) data_buf;
	strcpy(log_buf, "    ");
	for (i = 0; i < n; i++) {
	  sprintf(log_buf + 4 + (i % 16) * 3, "%02x ", *ptr++);
	  if ((i % 16) == 15 || i == n - 1)
	    els_log_msg(SYSCON|NOSTAMP, log_buf);
	}
#endif
	if (write(els_fd, data_buf, n) < 0)  {
	  els_log_err(SYSCON, "write on EtherLite:%s\n", strerror(errno));
	  err_exit();
	}
      } else {
	els_log_err(SYSCON, "read on socket:%s\n",strerror(errno));
	break;
      }
    }
  }
  close(sock_fd);

  els_log_msg(SYSCON, "Unit Offline\n");
  if (ioctl(els_fd, ELSOFFLINE, 0) < 0) {
    els_log_err(SYSCON, "ELSOFFLINE:%s\n",strerror(errno));
    err_exit();
  }
  goto main_loop;
}



int make_socket_connection()
{
  int inquiry_len;
  
  if ((sock_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)  {
    els_log_err(SYSCON, "socket:%s\n", strerror(errno));
    return(-1);
  }
#if 0
  {	int on = 1;
  setsockopt(sock_fd, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof(on));
  /*setsockopt(sock_fd, SOL_SOCKET, SO_LINGER, &on, sizeof(on));*/
  }
#endif
  bzero((char *)&cli_addr, sizeof(cli_addr));
  cli_addr.sin_family = AF_INET;
  cli_addr.sin_addr.s_addr = els_cfg.netaddr;
  cli_addr.sin_port = htons(ELS_TCP_PORT);

  if (connect(sock_fd, (struct sockaddr *) &cli_addr, 
	      sizeof(cli_addr)) < 0) {
    close(sock_fd);
    return(-1);
  }
  unlock_req.opcode = FAS_UNIT_UNLOCK;
  if (write(sock_fd, &unlock_req, FASCMDSZ) < 0) {
    els_log_err(SYSCON, "write FAS_UNIT_UNLOCK:%s\n",strerror(errno));
    close(sock_fd);
    return(-1);
  }
  if ((sock_read(sock_fd, (unsigned char *) &unlock_rsp, FASCMDSZ) < 0) ||
      unlock_rsp.opcode != FAS_UNIT_UNLOCK) {
    els_log_err(SYSCON,"read FAS_UNIT_UNLOCK:%s\n",strerror(errno));
    close(sock_fd);
    return(-1);
  }
  bitmash(unlock_rsp.challenge, inquiry_req.key);
  inquiry_req.opcode = FAS_UNIT_INQUIRY;
  if (write(sock_fd, &inquiry_req, FASCMDSZ) < 0) {
    els_log_err(SYSCON, "write FAS_UNIT_INQUIRY:%s\n", strerror(errno));
    close(sock_fd);
    return(-1);
  }
  if ((sock_read(sock_fd, (unsigned char *) &inquiry_rsp, FASCMDSZ) < 0) ||
      inquiry_rsp.opcode != FAS_UNIT_INQUIRY) {
    els_log_err(SYSCON, "read FAS_UNIT_INQUIRY:%s\n",strerror(errno));
    close(sock_fd);
    return(-1);
  }
  inquiry_len = inquiry_rsp.cnt_lo + (inquiry_rsp.cnt_hi << 8);
  if (sock_read(sock_fd, inquiry_data, inquiry_len) < 0) {
    els_log_err(SYSCON, "read inquiry data:%s\n",strerror(errno));
    close(sock_fd);
    return(-1);
  }
  return(0);
}


int sock_read(int sock_fd, unsigned char *buf, int wanted)
{
  int n;
  
  while (wanted)  {
    if ((n = read(sock_fd, buf, wanted)) <= 0) {
      return(-1);
    }
    buf += n;
    wanted -= n;
  }
  return(0);
}


/*
    bitmash function

    Accepts an 6 byte challenge and returns a 3 byte key
    which has been bitmashed in a moderately inscrutable way.
    Prevents inadvertant access of Etherlites.
*/
void
bitmash(challenge, key)
unsigned char *challenge;
unsigned char *key;
{
  int i, j, k, n, churns;
  
  /* mash the bits together */
  for( i=0, j=1, k=2, churns=0; churns < 31;
       i = (i + 1) % 6, j = (j + i) % 6, k = (k + j) % 6, churns++ ) {
    /* XOR a pair together */
    challenge[i] ^= challenge[j];
    /* rotate one byte each time */
    n = challenge[k] << 1;
    if( n & 0x100 )
      n |= 1;
    challenge[k] = (unsigned char) n;
  }
    /* build a key */
  key[0] = challenge[0] + challenge[4];
  key[1] = challenge[2] + challenge[1];
  key[2] = challenge[3] + challenge[5] + challenge[0];
}


void
els_terminate()
{
  err_exit();
}


void
els_log(msg, err_flag)
char *msg;
int err_flag;
{
  time_t cur_time;
  struct tm *tp;
  char buf[100];
  char err_buf[20];
  
  cur_time = time(0);
  tp = localtime(&cur_time);
  sprintf(buf, "%d/%d/%d %02d:%02d:%02d  %s.", tp->tm_mon+1,
	  tp->tm_mday, tp->tm_year, tp->tm_hour, tp->tm_min, tp->tm_sec, msg);
  
  if (err_flag) {
    sprintf(err_buf, " ERRNO %d\n", errno);
    strcat(buf, err_buf);
  }
  fprintf(stderr, "%s\n", buf);
}

void err_exit()
{
  els_log_msg(SYSCON, "Daemon process terminated\n");
  els_log_msg(SYSCON, "Unit offline\n");
  if(ioctl(els_fd, ELSOFFLINE, 0) < 0) 
    els_log_err(SYSCON, "ELSOFFLINE:%s\n",strerror(errno));
  close(els_fd);
  close(sock_fd);
  close(log_fd);
  exit(1);
}
